import enum

from datetime import date
from dateutil.utils import today

import numpy as np
from scipy.stats import norm


class Security:
    def __init__(self, symbol: str, price: float = None):
        self.symbol: str = symbol
        self.price: float | None = price


class Future(Security):
    def __init__(self, symbol, underlying: Security, **kwargs):
        super().__init__(symbol, **kwargs)

        self.underlying: Security = underlying


class ExerciseStyle(enum.Enum):
    american = "american"
    european = "european"
    asian = "asian"


class ExerciseType(enum.Enum):
    call = "put"
    put = "put"


class GenericOption:
    def __init__(
        self,
        security: Security,
        exercise_style: ExerciseStyle = ExerciseStyle.american,
        risk_free_rate: float = None,
        country: str = None,
        dividend_yield: float = None,
    ):
        """GenericOption to take care of risk_free_rate, dividend yield and correct pricing engine"""
        self.security = security
        self.exercise_style = exercise_style

        self.risk_free_rate: float | None
        self.dividend_yield: float | None = dividend_yield

        if risk_free_rate:
            self.risk_free_rate = risk_free_rate
        elif country:
            self.risk_free_rate = int(country)


class Option(Security):
    def __init__(
        self,
        symbol: str,
        underlying: GenericOption,
        strike: float,
        maturity: float | date,
        exercise_type: ExerciseType = ExerciseType.call,
        future: Security = None,
        **kwargs,
    ):
        super().__init__(symbol, **kwargs)

        self.underlying: GenericOption = underlying
        self.future: Security | None = future

        self.strike: float = strike
        self.maturity: float | date = maturity

        self.exercise_type = exercise_type

        self.value: float | None
        self.volatility: float | None

        self.__implied_value: float | None = None
        self.__implied_volatility: float | None = None

    def days_to_expire(self):
        if isinstance(self.maturity, date):
            return (self.maturity - today().date()).days
        else:
            return self.maturity * 360

    def pricing(self, implied_volatility, method=None):
        if method is None:
            if self.underlying.exercise_style == ExerciseStyle.european:
                return PricingEngine.black_scholes_merton(self, implied_volatility)

            elif self.underlying.exercise_style == ExerciseStyle.american:
                return PricingEngine.binomial_three(self, implied_volatility)

    @property
    def implied_value(self):
        return self.__implied_value

    @property
    def implied_volatility(self):
        return self.__implied_volatility

    @implied_volatility.setter
    def implied_volatility(self, implied_volatility):
        self.__implied_value = self.pricing(implied_volatility)
        self.__implied_volatility = implied_volatility

    def __str__(self):
        return f"{self.symbol};{self.maturity}-{self.strike}"


class PricingEngine:
    @classmethod
    def differentiate_greeks(cls, method=None):
        if method is None:
            method = cls.binomial_three
        # TODO: Implement differentiation
        #   delta <- function(x) {
        #     h <- 1e-2
        #     f <- function(x) {
        #     return(AmericanOption(type, x, strike, dividendYield, risk_free_rate, maturity, volatility)[["value"]]) }
        #     return((f(x + h) - f(x - h)) / (2 * h))
        #   }
        #
        #   gamma <- function(x) {
        #     h <- 1e-2
        #     f <- function(x) {
        #     return(AmericanOption(type, x, strike, dividendYield, risk_free_rate, maturity, volatility)[["value"]]) }
        #     return((f(x - h) - 2 * f(x) + f(x + h)) / (h^2))
        #   }
        #
        #   vega <- function(x) {
        #     h <- 5e-2
        #     f <- function(x) {
        #     return(AmericanOption(type, underlying, strike, dividendYield, risk_free_rate, maturity, x)[["value"]]) }
        #     return((f(x + 2 * h) - f(x - 2 * h)) / (4 * h))
        #   }
        #
        #   theta <- function(x) {
        #     h <- 1 / 365
        #     f <- function(x) {
        #     return(AmericanOption(type, underlying, strike, dividendYield, risk_free_rate, x, volatility)[["value"]]) }
        #     return((f(x - h) - f(x + h)) / (2 * h))
        #   }
        #
        #   rho <- function(x) {
        #     h <- 1e-2
        #     f <- function(x) {
        #     return(AmericanOption(type, underlying, strike, dividendYield, x, maturity, volatility)[["value"]]) }
        #     return((f(x + h) - f(x)) / h)
        #   }

    @classmethod
    def binomial_three(cls, option: Option, volatility: float):
        return volatility

    @classmethod
    def black_scholes_merton(cls, option: Option, volatility: float):
        underlying: GenericOption = option.underlying
        strike = option.strike
        maturity = option.days_to_expire() / 360

        risk_free_rate = underlying.risk_free_rate

        d_1 = np.log(underlying.security.price / strike) + (
            (np.power(volatility, 2) / 2) * maturity
        ) / (volatility * np.sqrt(maturity))
        d_2 = d_1 - (volatility * np.sqrt(maturity))

        if option.exercise_type == "call":
            price = np.exp(-risk_free_rate * maturity) * (
                underlying.security.price * norm.pdf(d_1) - strike * norm.pdf(d_2)
            )
        else:
            price = np.exp(-risk_free_rate * maturity) * (
                underlying.security.price * norm.pdf(-d_1) + strike * norm.pdf(-d_2)
            )

        def delta():
            if option.exercise_type == "call":
                return np.exp(-risk_free_rate * maturity) * norm.pdf(d_1)
            else:
                return -np.exp(-risk_free_rate * maturity) * norm.pdf(-d_1)

        def gamma():
            return (
                np.exp(-risk_free_rate * maturity)
                / (underlying.security.price * volatility * np.sqrt(maturity))
                * norm.cdf(d_1)
            )

        def vega():
            return (
                np.exp(-risk_free_rate * maturity)
                * norm.cdf(d_1)
                * underlying.security.price
                * np.sqrt(maturity)
            )

        def theta():
            if option.exercise_type == ExerciseType.call:
                return (
                    -(
                        underlying.security.price
                        * volatility
                        * np.exp(-risk_free_rate * maturity)
                        * norm.cdf(d_1)
                    )
                    / (2 * np.sqrt(maturity))
                    + risk_free_rate
                    * underlying.security.price
                    * np.exp(-risk_free_rate * maturity)
                    * norm.pdf(d_1)
                    - risk_free_rate
                    * strike
                    * np.exp(-risk_free_rate * maturity)
                    * norm.pdf(d_2)
                )
            else:
                return (
                    -(
                        underlying.security.price
                        * volatility
                        * np.exp(-risk_free_rate * maturity)
                        * norm.cdf(d_1)
                    )
                    / (2 * np.sqrt(maturity))
                    - risk_free_rate
                    * underlying.security.price
                    * np.exp(-risk_free_rate * maturity)
                    * norm.pdf(-d_1)
                    + risk_free_rate
                    * strike
                    * np.exp(-risk_free_rate * maturity)
                    * norm.pdf(-d_2)
                )

        def rho():
            if option.exercise_type == ExerciseType.call:
                return (
                    strike
                    * maturity
                    * np.exp(-risk_free_rate * maturity)
                    * norm.pdf(d_2)
                )
            else:
                return (
                    -strike
                    * maturity
                    * np.exp(-risk_free_rate * maturity)
                    * norm.pdf(-d_2)
                )

        # AmericanPricing <- function(type, underlying, strike, risk_free_rate, dividendYield, maturity, volatility) {
        #   value <- AmericanOption("call", underlying, strike, dividendYield, risk_free_rate, maturity, volatility)
        #   value <- c("price" = value[["value"]], "divRho" <- value[["divRho"]])
        #
        #   value["delta"] <- delta(underlying) * 1 # 1 change in underlying
        #   value["gamma"] <- gamma(underlying) * underlying # 1 change in underlying
        #   value["vega"] <- vega(volatility) / 100 # 1% change in volatility
        #   value["theta"] <- theta(maturity) / 365 # 1 day in year change
        #   value["rho"] <- rho(risk_free_rate) / (100) # 1bps curve change
        #
        value = {
            "price": price,
            "delta": delta(),
            "gamma": gamma(),
            "vega": vega(),
            "theta": theta(),
            "rho": rho(),
        }

        return value


def main():
    security = Security("PETR4 BR Index", price=40.5)

    underlying = GenericOption(
        security,
        risk_free_rate=10 / 100,
        dividend_yield=30 / 100,
        exercise_style=ExerciseStyle.european,
    )

    option = Option(
        "PETR4 BR 2023/7/20 C40",
        underlying,
        strike=40,
        maturity=date(2023, 7, 20),
        price=8,
    )

    option.implied_volatility = 4 / 100
    print(option.implied_value)


if __name__ == "__main__":
    main()
